import hashlib

from celery import Task
from celery.utils import uuid
from celery.utils.log import get_task_logger
from django.conf import settings
from django.core.cache import caches
from django.utils.functional import cached_property
try:
    from django_redis.cache import RedisCache
except ImportError:
    RedisCache = None

log = get_task_logger(__name__)


class CouldNotObtainLock(Exception):
    pass


class QueuedOnceTask(Task):
    abstract = True
    once_key_arg = None

    _LOCK_EXPIRE = 60 * 60 * 24  # 24 hours

    @cached_property
    def cache(self):
        backend = getattr(
            settings, 'CELERY_QUEUES_ONCE_CACHE_BACKEND', 'default')
        return caches[backend]

    def _key_from_args(self, args=None, kwargs=None):
        if args is None:
            args = ()
        if kwargs is None:
            kwargs = {}

        if self.once_key_arg is None:
            sorted_kwargs = [(key, kwargs[key]) for key in sorted(kwargs)]
            key_args = (args, sorted_kwargs)
        elif isinstance(self.once_key_arg, int):
            if len(args) <= self.once_key_arg:
                raise ValueError(
                    'Task requires at least {} positional argument(s) due to '
                    'once_key_arg={}'
                    .format(self.once_key_arg + 1, self.once_key_arg))
            key_args = args[self.once_key_arg]
        else:
            if self.once_key_arg not in kwargs:
                raise ValueError(
                    'Task requires keyword argument {0!r} due to once_key_arg'
                    .format(self.once_key_arg))
            key_args = kwargs[self.once_key_arg]

        unhashed = u"({},{},{})".format(
            self.__module__, self.__name__, key_args)
        hashed = hashlib.md5(unhashed.encode('utf-8'))
        return 'queuedtasks:{}'.format(hashed.hexdigest())

    def _get_lock(self, key):
        task_id = self.cache.get(key)
        return task_id

    def _take_lock(self, task_id, key):

        # Determine if we're using Redis as our cache
        is_redis = (RedisCache is not None and
                    isinstance(self.cache, RedisCache))

        # If we're using Redis, utilize SETNX
        set_kwargs = {'nx': True} if is_redis else {}

        # Set the lock in the cache
        success = self.cache.set(key, task_id, self._LOCK_EXPIRE, **set_kwargs)

        return success if is_redis else True

    def _clear_lock(self, key):
        self.cache.delete(key)

    @staticmethod
    def _propagates_exceptions():
        return getattr(settings, 'CELERY_EAGER_PROPAGATES_EXCEPTIONS', False)

    MAX_LOCK_SPINS = 40

    def apply_async(self, args=None, kwargs=None, **other):
        if self._propagates_exceptions():
            log.warning(
                'Cannot take a lock and reliably clear it when '
                'CELERY_EAGER_PROPAGATES_EXCEPTIONS is True.')
            return super(QueuedOnceTask, self).apply_async(
                args, kwargs, **other)

        # Determine lock cache key
        key = self._key_from_args(args, kwargs)

        # Generate or get the task_id
        new_task_id = other.setdefault('task_id', uuid())

        # Variable to represent if we've taken the lock
        lock_taken = False

        # Make sure we don't loop for too long
        spin_count = 0

        while not lock_taken:
            if spin_count > self.MAX_LOCK_SPINS:
                raise CouldNotObtainLock

            # See if this task is already queued
            existing_task_id = self._get_lock(key)
            if existing_task_id:
                log.debug(
                    'Got a duplicate task for one that was previously queued.',
                    extra={'data': {
                        'task_id': existing_task_id,
                        'name': self.__name__,
                        'args': args,
                        'kwargs': kwargs
                    }}
                )
                return self.AsyncResult(existing_task_id)

            lock_taken = self._take_lock(new_task_id, key)
            spin_count += 1

        # Actually apply the task
        return super(QueuedOnceTask, self).apply_async(args, kwargs, **other)

    def after_return(self, status, retval, task_id, args, kwargs, einfo):
        if not self._propagates_exceptions():
            self._clear_lock(self._key_from_args(args, kwargs))

        return super(QueuedOnceTask, self).after_return(
            status, retval, task_id, args, kwargs, einfo)



from celery import current_app
from celery.task import current, task
from django.conf import settings
from django.core.cache import get_cache
from django.test import SimpleTestCase
from django.test.utils import override_settings

from queued_once import CouldNotObtainLock, QueuedOnceTask


@task(base=QueuedOnceTask)
def lock_taken(*args, **kwargs):
    test_case = kwargs.get('test_case')
    lock = current._get_lock(current._key_from_args(args, kwargs))
    test_case.assertIsNotNone(lock)
    test_case.assertEqual(current.request.id, lock)


@task(base=QueuedOnceTask)
def recursive_task(*args, **kwargs):
    test_case = kwargs.get('test_case')
    test_case.assertEqual(0, recursive_task.count)
    recursive_task.count += 1
    result = recursive_task.delay(*args, **kwargs)
    test_case.assertEqual(result.id, current.request.id)
    test_case.assertEqual(1, recursive_task.count)


@task(base=QueuedOnceTask, once_key_arg='mykey')
def recursive_task_with_key(*args, **kwargs):
    test_case = kwargs.get('test_case')
    depth = kwargs.get('depth', 0)
    test_case.assertEqual(0, depth)
    kwargs['depth'] = depth + 1
    result = recursive_task_with_key.delay(*args, **kwargs)
    test_case.assertEqual(result.id, current.request.id)


@task(base=QueuedOnceTask)
def retry_task(*args, **kwargs):
    test_case = kwargs.get('test_case')
    retry_task.count += 1

    prev_count = retry_task.count
    retry_task.delay(*args, **kwargs)
    test_case.assertEqual(prev_count, retry_task.count)

    if retry_task.count < 3:
        return retry_task.retry(*args, **kwargs)
    test_case.assertEqual(current.request.retries, retry_task.count - 1)
    retry_task.delay(*args, **kwargs)
    test_case.assertEqual(3, retry_task.count)


class CustomException(Exception):
    pass


@task(base=QueuedOnceTask)
def exception_task(*args, **kwargs):
    raise CustomException('Custom')


@task(base=QueuedOnceTask)
def recursive_exception_task(*args, **kwargs):
    test_case = kwargs.get('test_case')
    recursive_exception_task.count += 1
    result = recursive_exception_task.delay(*args, **kwargs)
    test_case.assertEqual(result.id, current.request.id)
    test_case.assertEqual(1, recursive_exception_task.count)

    raise CustomException('Custom')


@override_settings(CELERY_EAGER_PROPAGATES_EXCEPTIONS=False)
class QueuedOnceTaskTest(SimpleTestCase):

    def setUp(self):
        backend = getattr(
            settings, 'CELERY_QUEUES_ONCE_CACHE_BACKEND', 'default')
        self.cache = get_cache(backend)
        try:
            self.cache.delete_pattern('queuedtasks:*')
        except AttributeError:
            self.cache.clear()
        current_app.config_from_object(settings)

    def assertLockNotTaken(self, task, *args, **kwargs):
        lock = task._get_lock(
            task._key_from_args(args, dict(test_case=self, **kwargs)))
        self.assertIsNone(lock)

    def assertIsUUID(self, value):
        self.assertEqual(36, len(value))

    def test_single_task(self):
        self.assertLockNotTaken(lock_taken)
        result = lock_taken.delay(test_case=self)
        result.get()
        self.assertIsUUID(result.id)
        self.assertLockNotTaken(lock_taken)

    def test_two_tasks(self):
        self.assertLockNotTaken(recursive_task)
        recursive_task.count = 0
        result = recursive_task.delay(test_case=self)
        result.get()
        self.assertIsUUID(result.id)
        self.assertLockNotTaken(recursive_task)

    def test_retry_task(self):
        self.assertLockNotTaken(retry_task)
        retry_task.count = 0
        result = retry_task.delay(test_case=self)
        result.get()
        self.assertIsUUID(result.id)
        self.assertLockNotTaken(retry_task)

    @override_settings(CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
    def test_exception_task(self):
        self.assertLockNotTaken(exception_task)
        with self.assertRaises(CustomException):
            exception_task.delay(test_case=self)
        self.assertLockNotTaken(exception_task)

    def test_recursive_exception_task(self):
        self.assertLockNotTaken(recursive_exception_task)
        recursive_exception_task.count = 0
        result = recursive_exception_task.delay(test_case=self)
        with self.assertRaises(CustomException):
            result.get()
        self.assertLockNotTaken(recursive_exception_task)
        self.assertEqual(1, recursive_exception_task.count)

    def test_could_not_obtain_lock(self):
        @task(base=QueuedOnceTask)
        def small_task(*args, **kwargs):
            pass

        small_task._take_lock = lambda x, y: False

        self.assertLockNotTaken(small_task)
        small_task.count = 0
        with self.assertRaises(CouldNotObtainLock):
            small_task.delay(test_case=self)
        self.assertLockNotTaken(small_task)
        self.assertEqual(0, small_task.count)

    def test_custom_key(self):
        self.assertLockNotTaken(recursive_task_with_key, mykey=42)
        recursive_task_with_key.delay(test_case=self, mykey=42).get()
        self.assertLockNotTaken(recursive_task_with_key, mykey=42)



#!/usr/bin/env python

from setuptools import setup


setup(
    name='celery-queued-once',
    version='0.2',
    description='Celery base task de-duplicating tasks',
    author='Educreations Engineering',
    packages=['queued_once'],
    install_requires=['celery', 'django >= 1.7'],
)
